// OBJ Export plugin.

#include <stdio.h>

#include "ac_plugin.h"

//---------------------------------------------------------------------------
class CStringList
{
public:
   CStringList *next;
   char *data;
   CStringList()
   {
      data = NULL;
      next = NULL;
   }
   static void freeList( CStringList **inlist )
   {
      CStringList *head = *inlist;
      while( head != NULL )
      {
         CStringList *delentry = head;
         head = head->next;
         delete delentry;
      }
   }
   static void addUnique( CStringList **inlist, char *value )
   {
      CStringList *head = *inlist;
      if( head == NULL )
      {
         head = new CStringList();
         *inlist = head;
         head->data = new char[strlen( value ) + 1];
         strcpy( head->data, value );
         return;
      }
      CStringList *tail = NULL;
      for( CStringList *p = head; p != NULL; p = p->next )
      {
         if( strcmp( p->data, value ) == 0 )
         {
            return;
         }
         tail = p;
      }
      tail->next = new CStringList();
      tail = tail->next;
      tail->data = new char[strlen( value ) + 1];
      strcpy( tail->data, value );
   }
};
//---------------------------------------------------------------------------
class COBJMaterial
{
public:
   COBJMaterial *next;
   int id;
   unsigned short materialIndex;
   char *textureName;
   COBJMaterial()
   {
      next = NULL;
      id = 0;
      textureName = NULL;
      materialIndex = 0;
   }
   ~COBJMaterial()
   {
      if( textureName != NULL )
      {
         delete textureName;
      }
   }
   int match( unsigned short col, char *tex )
   {
      for( COBJMaterial *p = this; p != NULL; p = p->next )
      {
         // If one but not both texture names are NULL, no match
         if(( p->textureName == NULL || tex == NULL ) && p->textureName != tex )
         {
            continue;
         }
         // Compare the material index and the texture names
         if(( p->materialIndex == col ) &&
            (( p->textureName == NULL && tex == NULL ) || strcmp( p->textureName, tex ) == 0 ))
         {
            return p->id;
         }
      }
      return -1;
   }
   void setTextureName( const char *t )
   {
      if( textureName != NULL )
      {
         delete textureName;
         textureName = NULL;
      }
      if( t != NULL )
      {
         textureName = new char[strlen( t ) + 1];
         strcpy( textureName, t );
      }
   }
   static int addMaterial( COBJMaterial **matlist, unsigned short col, char *tex )
   {
      COBJMaterial *head = *matlist;
      if( head == NULL )
      {
         head = new COBJMaterial();
         *matlist = head;
         head->id = 0;
         head->materialIndex = col;
         head->setTextureName( tex );
         return head->id;
      }
      int matchID = head->match( col, tex );
      if( matchID == -1 )
      {
         COBJMaterial *tail = NULL;
         for( COBJMaterial *p = head; p != NULL; p = p->next )
         {
            matchID = p->id + 1;
            tail = p;
         }
         tail->next = new COBJMaterial();
         tail = tail->next;
         tail->id = matchID;
         tail->materialIndex = col;
         tail->setTextureName( tex );
      }
      return matchID;
   }
   static void freeList( COBJMaterial **matlist )
   {
      COBJMaterial *head = *matlist;
      if( head == NULL )
      {
         return;
      }
      COBJMaterial *tail = NULL;
      while( head != NULL )
      {
         tail = head;
         head = head->next;
         delete tail;
      }
   }
};
//---------------------------------------------------------------------------
#define DDH_OBJ_MAXBUFCHARS 1024
class COBJData
{
public:
   char buffer[DDH_OBJ_MAXBUFCHARS + 1];
   unsigned int nextObject;
   unsigned int nextVertex;
   unsigned int nextSVertex;
   bool optWriteGroups;
   bool optUseAC3DGroups;
   bool optWriteFullTexturePaths;
   bool optCopyTextures;
   bool optBackslashTexturePaths;
   COBJMaterial *matList;
   int lastMatID;
   COBJData()
   {
      matList = NULL;
      lastMatID = -1;
      optWriteGroups = true;
      optUseAC3DGroups = true;
      optWriteFullTexturePaths = true;
      optBackslashTexturePaths = true;
      optCopyTextures = true;
      nextObject = 0;
      nextVertex = 1;
      nextSVertex = 1;
   }
};
//---------------------------------------------------------------------------
static void ddh_backslash_string( char *str )
{
   char *p = str;
   if( p == NULL )
   {
      return;
   }
   while( *p != '\0' )
   {
      if( *p == '/' )
      {
         *p = '\\';
      }
      ++p;
   }
}
//---------------------------------------------------------------------------
#define DDH_OBJ_COPYFILE_BUFFER 32768
static void copyFile( char *file1, char *file2 )
{
   if( strcmp( file1, file2 ) == 0 )
   {
      printf( "No need to copy %s - already in destination.\n", file1 );
      return;
   }
   printf( "Copying %s to %s\n", file1, file2 );
   FILE *fp1 = fopen( file1, "rb" );
   if( fp1 == NULL )
   {
      printf( "Error opening %s for reading\n", file1 );
      return;
   }
   FILE *fp2 = fopen( file2, "wb" );
   if( fp2 == NULL )
   {
      printf( "Error opening %s for writing\n", file2 );
      fclose( fp1 );
      return;
   }
   char *buffer = new char[DDH_OBJ_COPYFILE_BUFFER];
   int bytesRead;
   do
   {
      bytesRead = fread( buffer, sizeof( char ), DDH_OBJ_COPYFILE_BUFFER, fp1 );
      if( bytesRead != 0 )
      {
         int bytesWritten = fwrite( buffer, sizeof( char ), bytesRead, fp2 );
         if( bytesRead != bytesWritten )
         {
            printf( "Error copying %s - tried to write %d bytes, wrote %d bytes\n",
                     file1, bytesRead, bytesWritten );
            break;
         }
      }
   } while( bytesRead == DDH_OBJ_COPYFILE_BUFFER );
   fclose( fp1 );
   fclose( fp2 );
   delete[] buffer;
}
//---------------------------------------------------------------------------
static void printGroupID( FILE *fp, ACObject *ob, int groupID )
{
   // Check for an object name
   char *oname = ac_object_get_name( ob );
   if( oname == NULL || *oname == '\0' )
   {
      // Unnamed object
      fprintf( fp, "g g%d\n", groupID );
   }
   else
   {
      // Named object
      // Replace special characters with underscores
      char buffer[256];
      strncpy( buffer, oname, 255 );
      char *pchar = buffer;
      while( *pchar != '\0' )
      {
         if( *pchar < 33 )
         {
            *pchar = '_';
         }
         ++pchar;
      }
      // Output object name
      fprintf( fp, "g g%d_%s\n", groupID, buffer );
   }
}
//---------------------------------------------------------------------------
static void ddh_obj_output_object( FILE *fp, ACObject *ob, COBJData &objData )
{

   int numvert, numsurf, numkids;
   List *vertices, *surfaces, *kids;
   ac_object_get_contents( ob, &numvert, &numsurf, &numkids, &vertices, &surfaces, &kids );

   // Get the object's texture
   char *tex = NULL;
   if( ac_object_has_texture( ob ))
   {
      tex = texture_id_to_name( ac_object_get_texture_index( ob ));
   }

   // Write vertices
   List *p;
   unsigned int numVerts = 0;
   for( p = vertices; p != NULL; p = p->next )
   {
      Vertex *v = ( Vertex * )p->data;
      fprintf( fp, "v %g %g %g\n", v->x, v->y, v->z );
      ++numVerts;
   }

   // Write vertex normals
   for( p = surfaces; p != NULL; p = p->next )
   {
      Surface *s = ( Surface * )p->data;
      List *pv;
      for( pv = s->vertlist; pv != NULL; pv = pv->next )
      {
         SVertex *sv = ( SVertex * )pv->data;
         fprintf( fp, "vn %g %g %g\n", sv->normal.x, sv->normal.y, sv->normal.z );
      }
   }

   // Write vertex texture coords
   for( p = surfaces; p != NULL; p = p->next )
   {
      Surface *s = ( Surface * )p->data;
      List *pv;
      for( pv = s->vertlist; pv != NULL; pv = pv->next )
      {
         SVertex *sv = ( SVertex * )pv->data;
         fprintf( fp, "vt %g %g\n", sv->tx, sv->ty );
      }
   }

   // Write object ID
   if( objData.optWriteGroups )
   {
      if( !objData.optUseAC3DGroups )
      {
         printGroupID( fp, ob, objData.nextObject );
         ++( objData.nextObject );
      }
   }

   // Write faces
   unsigned int numSVerts = 0;
   for( p = surfaces; p != NULL; p = p->next )
   {
      Surface *s = ( Surface * )p->data;
      // Apply material
      int newMatID = COBJMaterial::addMaterial( &( objData.matList ), s->col, tex );
      if( newMatID != objData.lastMatID )
      {
         objData.lastMatID = newMatID;
         fprintf( fp, "usemtl acmat_%d\n", newMatID );
      }
      // Write the face/line data
      bool closedLine = false;
      switch( surface_get_type( s ) )
      {
      case SURFACE_CLOSEDLINE:
         closedLine = true;
      case SURFACE_LINE:
         fprintf( fp, "l" );
         break;
      default:
         fprintf( fp, "f" );
         break;
      }
      List *pv;
      unsigned int origsverts = numSVerts;
      for( pv = s->vertlist; pv != NULL; pv = pv->next )
      {
         SVertex *sv = ( SVertex * )pv->data;
         fprintf( fp, " %d/%d/%d",
                  list_index( vertices, SVERTEX( sv )) + objData.nextVertex,
                  objData.nextSVertex + numSVerts,
                  objData.nextSVertex + numSVerts );
         ++numSVerts;
      }
      // If this is a closed line, reprint the first vertex
      if( closedLine )
      {
         SVertex *sv = ( SVertex * )( s->vertlist->data );
         fprintf( fp, " %d/%d/%d",
                  list_index( vertices, SVERTEX( sv )) + objData.nextVertex,
                  objData.nextSVertex + origsverts,
                  objData.nextSVertex + origsverts );
      }
      fprintf( fp, "\n" );
   }

   // Increment vertex indices
   objData.nextVertex += numVerts;
   objData.nextSVertex += numSVerts;

}
//---------------------------------------------------------------------------
static void ddh_obj_output_group( FILE *fp, ACObject *ob, COBJData &objData )
{

   int obType = ac_object_get_type( ob );

   int thisObject = objData.nextObject;
   if( objData.optWriteGroups && objData.optUseAC3DGroups )
   {
      printGroupID( fp, ob, objData.nextObject );
      ++( objData.nextObject );
   }

   int numvert, numsurf, numkids;
   List *vertices, *surfaces, *kids;
   ac_object_get_contents( ob, &numvert, &numsurf, &numkids, &vertices, &surfaces, &kids );

   List *plist;
   bool printedSubGroup = false;
   for( plist = kids; plist != NULL; plist = plist->next )
   {
      ACObject *objkid = ( ACObject * )plist->data;
      if( ac_object_get_childrenlist( objkid ) == NULL )
      {
         if( ac_object_get_type( objkid ) == OBJECT_NORMAL )
         {
            if( printedSubGroup && objData.optWriteGroups && objData.optUseAC3DGroups )
            {
               printGroupID( fp, ob, objData.nextObject );
               printedSubGroup = false;
            }
            ddh_obj_output_object( fp, objkid, objData );
         }
      }
      else
      {
         ddh_obj_output_group( fp, objkid, objData );
         printedSubGroup = true;
      }
   }

}
//---------------------------------------------------------------------------
static void ddh_obj_output_materials( FILE *fp, COBJData &objData, char *outpath )
{
   // Create a container for a list of unique texture names, which will be populated
   // and acted upon only if the user chose to copy textures.
   CStringList *texlist = NULL;
   char dirsep;
   if( objData.optBackslashTexturePaths )
   {
      dirsep = '\\';
   }
   else
   {
      dirsep = '/';
   }
   // Prepare copy data
   char *copyDestination = NULL;
   int maxoutpath = 0;
   if( objData.optCopyTextures )
   {
      // Determine the largest buffer we'll need for creating an output path
      for( COBJMaterial *m = objData.matList; m != NULL; m = m->next )
      {
         if( m->textureName != NULL )
         {
            maxoutpath = max( maxoutpath, strlen( m->textureName ) + 1 );
         }
      }
      copyDestination = new char[maxoutpath + strlen( outpath )];
   }
   // Write the materials
   for( COBJMaterial *m = objData.matList; m != NULL; m = m->next )
   {
      fprintf( fp, "newmtl acmat_%d\n", m->id );
      ACMaterial *mat = ac_palette_get_material(( int )( m->materialIndex ));
      ACrgb rgb;
      float val;
      ACEntity *me = ( ACEntity * )mat;
      ac_entity_get_rgb_value( me, "diffuse", &rgb );
      fprintf( fp, "Kd %g %g %g\n", rgb.r, rgb.g, rgb.b );
      ac_entity_get_rgb_value( me, "ambient", &rgb );
      fprintf( fp, "Ka %g %g %g\n", rgb.r, rgb.g, rgb.b );
      ac_entity_get_rgb_value( me, "specular", &rgb );
      fprintf( fp, "Ks %g %g %g\n", rgb.r, rgb.g, rgb.b );
      ac_entity_get_float_value( me, "shininess", &val );
      fprintf( fp, "Ns %g\n", val );
      ac_entity_get_float_value( me, "transparency", &val );
      fprintf( fp, "Tr %g\n", val );
      if( m->textureName != NULL )
      {
         // Make sure we store the file to copy as "unbackslashed"
         ac_unbackslash_string( m->textureName );
         if( objData.optCopyTextures )
         {
            CStringList::addUnique( &texlist, m->textureName );
         }
         if( objData.optWriteFullTexturePaths )
         {
            if( objData.optCopyTextures )
            {
               strcpy( copyDestination, outpath );
               char *p = strrchr( m->textureName, '/' );
               if( p == NULL )
               {
                  strcat( copyDestination, m->textureName );
               }
               else
               {
                  ++p;
                  strcat( copyDestination, p );
               }
               if( objData.optBackslashTexturePaths )
               {
                  ddh_backslash_string( copyDestination );
               }
               fprintf( fp, "map_Kd %s\n", copyDestination );
            }
            else
            {
               if( objData.optBackslashTexturePaths )
               {
                  ddh_backslash_string( m->textureName );
               }
               fprintf( fp, "map_Kd %s\n", m->textureName );
            }
         }
         else
         {
            char *p = strrchr( m->textureName, '/' );
            if( p == NULL )
            {
               fprintf( fp, "map_Kd %s\n", m->textureName );
            }
            else
            {
               ++p;
               fprintf( fp, "map_Kd %s\n", p );
            }
         }
      }
   }
   if( objData.optCopyTextures )
   {
      // Copy each unique file
      for( CStringList *tl = texlist; tl != NULL; tl = tl->next )
      {
         strcpy( copyDestination, outpath );
         char *p = strrchr( tl->data, '/' );
         if( p == NULL )
         {
            strcat( copyDestination, tl->data );
         }
         else
         {
            ++p;
            strcat( copyDestination, p );
         }
         copyFile( tl->data, copyDestination );
      }
   }
   CStringList::freeList( &texlist );
   if( copyDestination != NULL )
   {
      delete[] copyDestination;
   }
}
//---------------------------------------------------------------------------
static void ddh_obj_save( int argc, char *argv[] )
{

   if( argc < 7 )
   {
      message_dialog( "Insufficient parameters passed to OBJ export" );
      return;
   }

   COBJData objData;
   objData.optWriteGroups           = ( strcmp( argv[1], "0" ) != 0 );
   objData.optUseAC3DGroups         = ( strcmp( argv[2], "0" ) != 0 );
   objData.optCopyTextures          = ( strcmp( argv[3], "0" ) != 0 );
   objData.optWriteFullTexturePaths = ( strcmp( argv[4], "0" ) != 0 );
   objData.optBackslashTexturePaths = ( strcmp( argv[5], "0" ) != 0 );

   // Save the config data
   FILE *fp = fopen( argv[6], "w" );
   if( fp == NULL )
   {
      message_dialog( "Error saving options to %s", argv[6] );
   }
   else
   {
      fwrite( argv[1], 1, 1, fp );
      fwrite( argv[2], 1, 1, fp );
      fwrite( argv[3], 1, 1, fp );
      fwrite( argv[4], 1, 1, fp );
      fwrite( argv[5], 1, 1, fp );
      fclose( fp );
   }

   char *filter[] = { "OBJ files", ".obj", "All files", "*", NULL };
   char *fname = ac_get_save_filename( "Export an OBJ file", filter );
   if(( fname == NULL ) || ( *fname =='\0' ))
   {
      return;
   }
   fp = fopen( fname, "w" );
   if( fp == NULL )
   {
      message_dialog( "can't open file '%s' for writing", fname );
      return;
   }

   // Create the material filename - replace the existing OBJ extension with MTL
   char *newfilename = new char[strlen( fname ) + 4];
   strcpy( newfilename, fname );
   char *p = strrchr( newfilename, '.' );
   if( p != NULL )
   {
      *p = '\0';
   }
   strcat( newfilename, ".mtl" );

   // Print the filename as a material library
   ac_unbackslash_string( newfilename );
   p = strrchr( newfilename, '/' );
   if( p == NULL )
   {
      fprintf( fp, "mtllib %s\n", newfilename );
   }
   else
   {
      ++p;
      fprintf( fp, "mtllib %s\n", p );
      --p;
   }

   ddh_obj_output_group( fp, ac_get_world(), objData );
   fclose( fp );

   fp = fopen( newfilename, "w" );
   if( fp == NULL )
   {
      message_dialog( "can't open material file '%s' for writing", newfilename );
      COBJMaterial::freeList( &( objData.matList ));
      return;
   }
   // Write the materials
   // Get the path-only for newfilename
   p = strrchr( newfilename, '/' );
   if( p == NULL )
   {
      ddh_obj_output_materials( fp, objData, "." );
   }
   else
   {
      ++p;
      *p = '\0';
      ddh_obj_output_materials( fp, objData, newfilename );
   }
   delete[] newfilename;
   fclose( fp );
   COBJMaterial::freeList( &( objData.matList ));

}
//---------------------------------------------------------------------------
static void ddh_obj_showdialog( int argc, char *argv[] )
{
   tcl_command( "wfobj_dialog" );
   if( argc < 2 )
   {
      return;
   }
   char *fname = new char[strlen( argv[1] ) + 32];
   sprintf( fname, "%s/wf_obj_export.cfg", argv[1] );
   tcl_command( "set ddh_obj_cfgfile \"%s\"", fname );
   FILE *fp = fopen( fname, "r" );
   delete[] fname;
   if( fp == NULL )
   {
      return;
   }
   char opts[5];
   int ct = fread( opts, 1, 5, fp );
   fclose( fp );
   if( ct != 5 )
   {
      return;
   }
   tcl_command( "set ddh_obj_writegroups \"%c\"", opts[0] );
   tcl_command( "set ddh_obj_ac3dgroups \"%c\"", opts[1] );
   tcl_command( "set ddh_obj_copytextures \"%c\"", opts[2] );
   tcl_command( "set ddh_obj_fulltexpaths \"%c\"", opts[3] );
   tcl_command( "set ddh_obj_backslashpaths \"%c\"", opts[4] );
}
//---------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/wf_obj_export.tcl\"", path );
      ac_add_command_full( "ddh_obj_export", ddh_obj_save, 6, "argv",
                           "<writegroups> <ac3dgroups <copytex> <fullpaths> <backslashpaths> <configfile>",
                           "Enhanced Wavefront OBJ export" );

      ac_add_command_full( "ddh_obj_dialog", ddh_obj_showdialog, 1, "argv",
                           "<path to plugin dir>", "Shows enhanced OBJ export dialog" );
      //ac_add_export_menu_item( "Wavefront OBJ (enhanced)...", "wfobj_dialog");
      char *tcommand = new char[strlen( path ) + 32];
      sprintf( tcommand, "ac3d ddh_obj_dialog {%s}", path );
      ac_add_export_menu_item( "Wavefront OBJ (enhanced)...", tcommand );
      delete[] tcommand;

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Wavefront OBJ (enhanced) export plugin - Version 1.0 by Dennis Hawthorne" );
}
//---------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Wavefront OBJ (enhanced) exporter with save options" );
}
//---------------------------------------------------------------------------

